// FLASHUP - EPC System Update Utility (c) 2000 Technologic Systems
//
// Description
// --------------------------
// This program is used to load the onboard Flash, CMOS memory, and RTC 
// values for Technologic Systems EPCs. If the EPC is virgin (never loaded)
// or dead (the BIOS has been trashed), a TS-9100 or TS-9200 board must be 
// used to bootstrap the system. If the EPC is currently functional and
// you are simply updating some or all of the system, then this utility and
// the update image files can be downloaded to the EPC and run directly 
// from there.
//
// Flashup is also used to create the image files from a live EPC, enabling
// you to clone EPC installations.
//
// Flashup can currently handle the following sections:
//      BIOS	- create / restore the BIOS
//	DOS	- create / restore the DOS 
//	DISK	- create / restore the SSD (solid state disk)
//	CMOS	- create / restore the CMOS settings
//	RTC	- copy the time and date from a TS-9200 boot board
//
// Creating Update Images
// --------------------------
// To create the update images, the following syntax is used:
//
// FLASHUP /CREATE (BIOS(=F1)) (DOS(=F2)) (DISK(=F3)) (CMOS(=F4))
//
// Any combination of BIOS, DOS, DISK, and CMOS can be specified.
// Filenames can be specified for each of the sections, but are 
// optional. If a section is specified without a filename, it defaults
// to 
//	<4 digit model number><section name>.BIN
//
// For example, default filenames for a TS-2200 are 2200BIOS.BIN, 
// 2200DOS.BIN, 2200DISK.BIN, and 2200CMOS.BIN.
//
// When creating the images, Flashup will NOT overwrite a file that 
// already exists. You must delete, rename, or move the old copy first.
//
// The procedure for creating update images is as follows:
// 1.	Configure a master EPC from which the clones will be copied. In 
//	particular, make sure the Flash disk contains all necessary files, 
//	and the CMOS boot options are correct.
// 2.	Install the TS-9100 / TS-9200 boot board and power up the EPC.
// 3.	Change to the directory where the update images will be stored. By
//	default, this is 'C:\IMAGES'.
// 4.	If you are replacing a previous set of images, rename or move the
//	old copies of the files. Flashup will not overwrite files.
// 4.	Run flashup, specifying the sections for which you are creating 
//	images. Specify filenames if you don't want to use the defaults.
// 5.	Power off, remove the boot board (if present) and you are done.
//
// Updating an EPC (BLAST)
// --------------------------
// To update an EPC from image files, the following syntax is used:
//
// FLASHUP /BLAST (BIOS(=F1)) (DOS(=F2)) (DISK(=F3)) (CMOS(=F4)) (RTC)
//
// Any combination of BIOS, DOS, DISK, CMOS, and RTC can be specified.
// Filenames can be specified for each of the sections, but are 
// optional. If a section is specified without a filename, it defaults
// to 
//	<4 digit model number><section name>.BIN
//
// For example, default filenames for a TS-2200 are 2200BIOS.BIN, 
// 2200DOS.BIN, 2200DISK.BIN, and 2200CMOS.BIN.
//
// The RTC option copies the time and date from an attached TS-9200 boot
// board, which contains its own RTC. Without a TS-9200, you cannot use
// this feature.
//
// The procedure for updating an EPC is as follows:
// 1.	If this is a virgin board or one on which the BIOS has been 
//	corrupted (the red LED never even blinks at poweron) then you 
//	must use a TS-9100 or TS-9200 to bootstrap the EPC to a running
//	state. Live boards can be updated directly by downloading images
//	and Flashup to the EPC and running them. In either case, Flashup
//	and the update images must be accessable from the EPC.
// 3.	Change to the directory where the update images are stored. By
//	default, this is 'C:\IMAGES'.
// 4.	Run flashup, specifying the sections you are updating. Specify
//	filenames if you are not using the defaults.
// 5.	Power off, remove the boot board (if present) and you are done.

// Revision History
// --------------------------
// 19990705 1.0	cjd	Started
// 19990901 1.1	cjd	Modified and updated to read Flash image from a file.
//			Renamed to FLASHUP
// 19990905 1.2	cjd	Added ability to write image of Flash disk to a file.
// 19990914 1.3	cjd	Added ability to selectively create or restore Flash 
//			image files
// 19990915 1.3.1	Handles 'PS' returned by Prosoft PS-4100 BIOS call
// 20000501 1.4	cjd	Cmdline improved. Filenames can now be specified 
//			on the cmdline, as well as a prefix
// 20000616 1.5	cjd	Simplified cmdline - no prefix crap.
// 20000713 1.6	cjd	Added TS-2250
// 20000718 1.7 cjd	Added CMOS and RTC options, updated docs
// 20000727 1.7.1 cjd	- Dont quit if RTC set is attempted w/o a TS9100 - so
//			we can use the same batch files on a TS9100
//			- Give better feedback (what's going to be replaced)
//			for the confirm mesg for /BLAST and /EPROM
// 20001011 1.7.2 cjd	Added /BLASTBOOT flag to write to 91/9200 flash. This
//			is VERYVERY dangerous, so it is not in the help messages..
// 20010215 1.7.3 mtj Added RTC blast for the 9300
// 20010223 1.7.4 mtj changed RTC on 9200/9300 to 0x9E 0x9F so address is 1 byte

const char VER[] = "EPC System Update Utility v1.7.4, (c) 2001 Technologic Systems, Inc.";


//#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include <iostream.h>
#include <iomanip.h>
#include <string.h>
#include <sys\stat.h>


// CMOS defines
const int   CMOS_ADDR      = 0x70;
const int   CMOS_DATA      = 0x71;
const int   CMOS_ADDR_9200       = 0x9E;
const int   CMOS_DATA_9200       = 0x9F;

const int   CMOS_YEAR      = 0x09;
const int   CMOS_REGA      = 0x0A;
const int   CMOS_REGB      = 0x0B;
const int   CMOS_REGC      = 0x0C;
const int   CMOS_REGD      = 0x0D;
const int   CMOS_STAT      = 0x0E;
const int   CMOS_RESTART      = 0x0F;	// restart status byte	
const int   CMOS_BEGIN     = 0x10;	// start of cmos area	

const int   CMOS_ARRAY_SIZE      = 0x80;


// Flash BIOS commands
const int LOCK = 00;
const int ERASE = 01;
const int READ = 02;
const int WRITE = 03;

//const unsigned long FLASHBASE_SOURCE = 0x20000000;
const unsigned long FLASHBASE    = 0x10000000;
const unsigned long EEPROMBASE      = 0x20000000;
const unsigned long BOOTBASE     = 0x30000000;

const unsigned long BIOSOFFSET   = 0x000F0000;
const unsigned long DOSOFFSET    = 0x000E0000;
const unsigned long DISKOFFSET   = 0x00000000;

const unsigned long BOOTBIOSOFFSET     = 0x00070000;
const unsigned long BOOTDOSOFFSET      = 0x00060000;
const unsigned long BOOTDISKOFFSET     = 0x00000000;

const unsigned long BUFFER_SIZE  = 0x00004000;	// 16k buffer for data

#define CF 1  /* Carry flag */


// return codes
enum ErrorCodes
	{
	ERR_NONE = 0, ERR_USAGE, ERR_FILE, ERR_FLASH, ERR_NO9200
	};

const char *ErrorMesg[] = {
	"No Errors.",
	"Usage error (bad command line, or answered 'n' to erasing Flash).",
	"File read/write error.",
	"Flash read/write/erase error.",
	"Attempting to copy RTC. You must have a TS-9200 or TS-9300 for this option."};

// global variables that would be more trouble than it's worth to pass around everywhere..
//
// {none so far}


// Print some help based on what the error was, then exit.
void ErrorHandler (const int ErrNum) 
{
	cout << "Houston, we have a problem: " << ErrorMesg[ErrNum] << endl;
	exit(ErrNum);
}

// Display usage of the program and exit
void Usage ()
{
	cout << "Usage:\nFLASHUP (/BLAST|/CREATE) (BIOS(=F1)) (DOS(=F2)) (DISK(=F3)) (CMOS(=F4)) (RTC)\n\n";

	cout << "Choose one of the two commands:" << "\n";
	cout << "\t/BLAST\tWrite image files to Flash memory." << "\n";
	cout << "\t/CREATE\tRead Flash memory and write to image files." << "\n\n";

	cout << "Choose any combination of the following options:" << "\n";
	cout << "\tBIOS\tCreate/restore BIOS image (64KB @ 0x100F0000)" << "\n";
	cout << "\tDOS\tCreate/restore DOS image  (64KB @ 0x100E0000)" << "\n";
	cout << "\tDISK\tCreate/restore DISK image (896KB @ 0x10000000)" << "\n";
	cout << "\tCMOS\tCreate/restore CMOS image" << "\n";
	cout << "\tRTC\tCopy time and date from a TS-9200 or TS-9300 to the EPC RTC" << "\n\n";

	cout << "Optional Parameters" << "\n";
	cout << "\tF1" << "\n";
	cout << "\tF2\tThe Flash image filenames can be specified directly.\n";
	cout << "\tF3\n";
	cout << "\tF4\n\n";


	cout << "NOTE: When using /CREATE, Filename(s) must NOT exist - FLASHUP will NOT\n";
	cout << "      overwrite existing files." << "\n";

	cout << endl;

	exit (ERR_USAGE);
}

// Check whether we are running on a TS EPC
int CheckTSBios ()
{
	union REGS regs;

	regs.x.ax = 0xB000;		// TS BIOS info function
	int86 (0x15, &regs, &regs);	  // do it.

	// 'TS' should be returned in DX (the PS-4100 returns PS, so handle that too..)
	if(regs.h.ah == 0 && ((regs.x.dx == 0x5453) || (regs.x.dx == 0x5053)))
		return(regs.h.cl);  // YES - return the model number
	else
		return(0);	  // NO this is NOT TS BIOS
} // CheckTSBios

//
// RTC stuff
//

// unsigned char ReadCmosByte(int Base, unsigned char Address)
// 
// 	Read a byte of data from a RTC.
//
//	Base	- base i/o address of the RTC to read from
//	Address	- RTC address to read
//
unsigned char ReadCmosByte(int Base, unsigned char Address)
{
	unsigned char  tmp = 0;
	int   i = 0;

	asm   cli;			// turn off interrupts

	if(Address <= CMOS_YEAR)
		{
		// if the RTC is updating, wait until it is finished
		for(outportb(Base, CMOS_REGA); (inportb(Base + 1) & 0x80); outportb(Base, CMOS_REGA))
			{
			i++;
			}
		}

	outportb(Base, Address);
	tmp = inportb(Base + 1);

	if(Address <= CMOS_YEAR)
		{
		// if the RTC started updating while we were reading, wait until it is finished and try again
		for(outportb(Base, CMOS_REGA); (inportb(Base + 1) & 0x80); outportb(Base, CMOS_REGA))
			{
			i++;
			}

		outportb(Base, Address);
		tmp = inportb(Base + 1);
		}

	asm   sti;			// turn on interrupts
	return(tmp);
}

// unsigned char ReadCmosData()
// 
//	Reads bytes from Cmos into an array
//
//	Base		base i/o address of the RTC to read
// 	CmosData	array to store data in
//	Address		RTC address at which to start 
//	Length		number of bytes to read
//
//
void ReadCmosData(int Base, unsigned char CmosData[], unsigned char Address, unsigned char Length)
{
	unsigned char  i;

	cout << "Reading " << hex << (int)Length << "h bytes at cmos address " << (int)Address << "h\n";

	for(i = Address; i < (Address + Length); i++)
		{
		CmosData[i] = ReadCmosByte(Base, i);
		}
}

// unsigned char WriteCmosByte(int Address, unsigned char Data)
// 
// 	writes the data into RTC CMOS memory at Address
//
void WriteCmosByte(int Base, int Address, unsigned char Data)
{
	asm   cli;

	// if this is a RTC register, turn off updates while we are in here
	if(Address <= CMOS_YEAR)
		{
		outportb(Base, CMOS_REGB);
		outportb(Base + 1, (inportb(Base + 1) | 0x80));
		}

	outportb(Base, Address);
	outportb(Base + 1, Data);

	// if this is a RTC register, turn updates back on 
	if(Address <= CMOS_YEAR)
		{
		outportb(Base, CMOS_REGB);
		outportb(Base + 1, (inportb(Base + 1) & 0x7F));
		}

	asm   sti;
}

// unsigned char WriteCmosData()
// 
//	Reads bytes from an array into Cmos
//
//	Base		base i/o address of RTC
// 	CmosData	array to store data in
//	Address		RTC address at which to start 
//	Length		number of bytes to read
//
//
void WriteCmosData(int Base, unsigned char CmosData[], unsigned char Address, unsigned char Length)
{
	unsigned char  i;

	cout << "Writing " << hex << (int)Length << "h bytes at cmos address " << (int)Address << "h\n";
	for(i = Address; i < (Address + Length); i++)
		{
		WriteCmosByte(Base, i, CmosData[i]);
		}
}

// dumps the contents of the CMOS and RTC to the console
int DumpCmos(unsigned char CmosData[])
{
	int   i;		// loop control

	cout << "Dumping CMOS Array:\n" << hex;
	for(i = 0; i < CMOS_ARRAY_SIZE; i++)
		{
		cout << setw(4) << (int)CmosData[i];
		if((i + 1) % 16 == 0)
			cout << "\n";

		}
	cout << endl;
	return(0);
}

// Error = WriteCmosToFile(FileCmos, CmosData);
//
//	Filename	= File to write cmos image to
//	CmosData	= array of cmos data
//
// Returns 0 if no error, non-zero if any errors
int WriteCmosToFile(const char *Filename, unsigned char CmosData[])
{

	int   out_file;		// output file descriptor
	unsigned long  BytesWritten;	 // number of bytes written to the file

	out_file = open(Filename, O_WRONLY | O_CREAT | O_EXCL | O_BINARY, S_IWRITE);	// attempt to open new Filename - error if it already exists 
	if(out_file < 0)
		{
		cerr << "Error: Cannot open " << Filename << ". Does it already exist?" << endl;
		return(ERR_FILE);
		}

	ReadCmosData(CMOS_ADDR, CmosData, CMOS_BEGIN, (CMOS_ARRAY_SIZE - CMOS_BEGIN));

	if((BytesWritten = write(out_file, CmosData, CMOS_ARRAY_SIZE)) != CMOS_ARRAY_SIZE)
		{
		cerr << "Error: cannot write to file (disk full?)" << endl;
		return(ERR_FILE);
		}

	// All finished. Close things down and report success
	cout << "\rReading Cmos memory: complete                              " << endl; 
	close(out_file);
	cout << "Flash image file successfully updated. Total bytes = 0x" << hex << BytesWritten << endl;

	return(ERR_NONE);						 // successful exit
}     

// Error = WriteFileToCmos(FileCmos, CmosData);
//
//	Filename	= File to write cmos image to
//	CmosData	= array of cmos data
//
// Returns 0 if no error, non-zero if any errors
int WriteFileToCmos(const char *Filename, unsigned char CmosData[])
{

	int   in_file;		// input file descriptor
	unsigned long  read_size = 0;	// bytes read in last read

	in_file = open(Filename, O_RDONLY | O_BINARY);	// open the input file, binary read only
	if(in_file < 0)
		{
		cerr << "Error: cannot open " << Filename << endl;
		return(ERR_FILE);
		}

	if((read_size = read(in_file, CmosData, CMOS_ARRAY_SIZE)) != CMOS_ARRAY_SIZE)
		{
		cerr << "Error: cannot read from file (is it empty?)" << endl;
		return(ERR_FILE);
		}

	WriteCmosData(CMOS_ADDR, CmosData, CMOS_BEGIN, 0x40 - CMOS_BEGIN);

	// All finished. Close things down and report success
	cout << "\rWriting Cmos memory: complete                              " << endl; 
	close(in_file);
	cout << "Cmos memory successfully updated. Total bytes = 0x" << hex << read_size << endl;

	return(ERR_NONE);						 // successful exit
}     

// Error = UpdateRtc();
//
// Update time and date registers from TS-9200 RTC to the base board
//
// Checks that a TS-9200 is present
// Returns 0 if no error, non-zero if any errors
//
// the 9200 has 0x04 programmed at address 0x475, the 9300 has 0x32
//
// mtj 20010214 the 9100 is now obsolete, the 9300 has an RTC, so allow for 
//              9200 or 9300

int UpdateRtc()
{

	unsigned char  RtcData[CMOS_ARRAY_SIZE] = {0xFF};

	cout << "Checking for presence of TS-9200 or TS-9300... ";
	if(((inportb(0x475) & 0x07) != 0x04) && (inportb(0x475) != 0x32))
		{
		cout << "not found. Skipping RTC set.\n";
		return(ERR_NO9200);
		}
	else
		{
		cout << "found aok.\n";
		}

	// get RTC data from TS-9200
	ReadCmosData(CMOS_ADDR_9200, RtcData, 0, 0x0B);

	// copy to base board
	WriteCmosData(CMOS_ADDR, RtcData, 0, 0x0B);

	// All finished. Close things down and report success
	cout << "\rWriting RTC: complete                              " << endl; 
	cout << "RTC successfully updated." << endl;

	return(ERR_NONE);						 // successful exit
}     


//-----------------------------
//
// Flash stuff
//

// run a BIOS Flash command - erase / read / write
int FlashCmd (int Cmd, unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count)
{
	struct REGPACK reg;

	reg.r_ax = ((0xE0 << 8) | (Cmd & 0x00FF));	// shift E0h into  AH, put Cmd in AL

	reg.r_di = FlashAddr >> 16;		// shift FlashAddr MSW into DI
	reg.r_si = FlashAddr & 0x0000FFFF;	// Mask FlashAddr LSW, place in si
	reg.r_cx = Count;			// Load number of bytes to work on
	reg.r_es = BufSeg;
	reg.r_bx = BufOff;


//	cout << "\nFlashCmd: cmd=" << reg.r_ax << " Flashaddr=" << reg.r_di << ":" << reg.r_si 
//			<< " cnt=" << reg.r_cx << " Bufaddr=" << reg.r_es << ":" << reg.r_bx 
//			<< " flags=" << reg.r_flags << endl;

	intr(0x15, &reg);			// do it	

//	cout << "\nFlashCmd: cmd=" << reg.r_ax << " Flashaddr=" << reg.r_di << ":" << reg.r_si 
//			<< " cnt=" << reg.r_cx << " Bufaddr=" << reg.r_es << ":" << reg.r_bx 
//			<< " flags=" << reg.r_flags << endl;

	if(reg.r_flags & CF)
		{
//		cout << "Flashcmd: Returning ERROR!\n";
		return(1);	  // Flash command failure	
		}
	else
		{
//		cout << "Flashcmd: Returning AOK!\n";		
		return(0);	  // Flash command success
		}

} 


int ReadFlash (unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count)
{
//	return (FlashCmd (READ, FlashAddr, BufSeg, BufOff, Count));
	FlashCmd (READ, FlashAddr, BufSeg, BufOff, Count);
	return(0);							  // Always get errors with BIOS call - WHY?

}

int WriteFlash (unsigned long FlashAddr, unsigned int BufSeg, unsigned int BufOff, unsigned int Count)
{
//	// saftey check - prevents us from accidentally erasing the DOS or BIOS images.
//	if ((FlashAddr + Count - 1) >= 0x100E0000)
//		return(2);
//	else
	return(FlashCmd (WRITE, FlashAddr, BufSeg, BufOff, Count));
}

int EraseFlash (unsigned long FlashAddr)
{
//	// saftey check - prevents us from accidentally erasing the DOS or BIOS images.
//	if (FlashAddr >= 0x100E0000)
//		return(2);
//	else
	return(FlashCmd (ERASE, FlashAddr, 0, 0, 0));
}

int CheckModel()
{
	int   Model;

	cout << "Model number: ";
	Model = CheckTSBios();
	cout << "[" << Model << "] ";
	switch(Model)
		{
		case 0:
			cout << "Not a TS EPC! This PC is not supported - exiting." << endl;
			exit (ERR_USAGE);
			break;
		case 21:
			cout << "TS-2100 EPC" << endl; return(2100);
			//break;
		case 22:
			cout << "TS-2200 EPC" << endl; return(2200);
			//break;
		case 23:
			cout << "TS-2350 EPC" << endl; return(2350);
			//break;
		case 24:
			cout << "TS-2250 EPC" << endl; return(2250);
			//break;

		case 28:
			cout << "TS-2800 EPC" << endl; return(2800);
			//break;
		case 41:
			cout << "Prosoft Technology PS-4100 EPC" << endl; return(4100);
			//break;

		default:
			cout << "Unknown TS EPC model! (" << Model << ") - continuing" << endl; return(Model);
			//break;
		}
	return(Model);
}

// int WriteFileToFlash(const char *Filename, const unsigned long FlashAddress, const unsigned long FlashLength)
// 
// Erase a section of Flash and replace with the contents of Filename
//
//	Filename	= File containing bytes to write to Flash
//	FlashAddress	= Flash address to start from
//	FlashLength	= Number of bytes to write to Flash from the file
//
// Returns 0 if no error, non-zero if any errors
//
int WriteFileToFlash(const char *Filename, const unsigned long FlashAddress, const unsigned long FlashLength)
{
	char  Buffer[BUFFER_SIZE];	// data buffer
//	int 	i = 0;			// loop count
	int   in_file;		// input file descriptor
	unsigned long  read_size = 0;	// bytes read in last read
	unsigned long  read_total = 0;	// running total of bytes
	unsigned long  FlashTemp = 0;	// various temporary large numbers


	in_file = open(Filename, O_RDONLY | O_BINARY);	// open the input file, binary read only
	if(in_file < 0)
		{
		cerr << "Error: cannot open " << Filename << endl;
		return(ERR_FILE);
		}
	FlashTemp = filelength(in_file);	 // small sanity check here 
//	cout << "Image file size: 0x" << hex << FlashTemp << endl;
	if(FlashTemp < FlashLength)
		{
		cout << "Warning: Image file is smaller than intended Flash memory section. Continuing.\n";
		}

	// erase the Flash memory
	for(FlashTemp = FlashAddress; FlashTemp < (FlashAddress + FlashLength); FlashTemp += 0x10000)
		{

		cout << "\rErasing Flash memory: Address = 0x" << hex << FlashTemp;

		if((EraseFlash (FlashTemp)) != 0)
			{
			cout << endl << endl << "Flash memory erase command failure, exiting." << endl;
			return(ERR_FLASH);
			}
		}
	cout << "\rErasing Flash memory: complete                           " << endl;

	// write the new Flash contents
	read_total = 0;

	for(FlashTemp = FlashAddress; 
		(read_size = read(in_file, Buffer, sizeof(Buffer))) != 0; 
		FlashTemp += read_size)
		{

		cout << "\rWriting Flash memory: address = 0x" << hex << FlashTemp
		<< ", Bytes = 0x" << hex << read_size;

		if((WriteFlash(FlashTemp, FP_SEG(Buffer), FP_OFF(Buffer), read_size)) != 0)
			{
			cout << endl << endl << "Flash memory write command failure, exiting." << endl;
			return(ERR_FLASH);
			}
		read_total += read_size;  // increment bytes read count
		}

	// All finished. Close things down and report success
	close(in_file);
	cout << "\rWriting Flash memory: Done. Total bytes updated = 0x" << hex << read_total << endl;

	return(ERR_NONE);						 // successful exit
}

// int WriteFlashToFile(const char *Filename, const unsigned long FlashAddress, const unsigned long FlashLength)
// Read a section of Flash and write image to filename
//
//	Filename	= File to write Flash image to
//	FlashAddress	= Flash address to start from
//	FlashLength	= Number of bytes to write to the image file from Flash
//
// Returns 0 if no error, non-zero if any errors
//
int WriteFlashToFile(const char *Filename, const unsigned long FlashAddress, const unsigned long FlashLength)
{
	char  Buffer[BUFFER_SIZE];	// data buffer
//	int 	i = 0;			// loop count
	int   out_file;		// output file descriptor
	unsigned long  BytesWritten;	 // number of bytes written to the file
	unsigned long  BytesTotal = 0;	// total bytes written to the file
	unsigned long  FlashTemp = 0;	// various temp. large numbers

	out_file = open(Filename, O_WRONLY | O_CREAT | O_EXCL | O_BINARY, S_IWRITE);	// attempt to open new Filename - error if it already exists 
	if(out_file < 0)
		{
		cerr << "Error: Cannot open " << Filename << ". Does it already exist?" << endl;
		return(ERR_FILE);
		}

	for(FlashTemp = FlashAddress; FlashTemp < (FlashAddress + FlashLength); FlashTemp += BUFFER_SIZE)
		{
		cout << "\rReading Flash memory: address = 0x" << hex << FlashTemp
		<< ", Bytes = 0x" << hex << BUFFER_SIZE;

		if(ReadFlash (FlashTemp, FP_SEG(Buffer), FP_OFF(Buffer), BUFFER_SIZE))
			{
			cerr << "Error: cannot read from Flash memory - Unknown Error" << endl;
			return(ERR_FLASH);
			}
		if((BytesWritten = write(out_file, Buffer, BUFFER_SIZE)) != BUFFER_SIZE)
			{
			cerr << "Error: cannot write to file (disk full?)" << endl;
			return(ERR_FILE);
			}
		BytesTotal += BytesWritten;
		}

	// All finished. Close things down and report success
	cout << "\rReading Flash memory: complete                              " << endl; 
	close(out_file);
	cout << "Flash image file successfully updated. Total bytes = 0x" << hex << BytesTotal << endl;

	return(ERR_NONE);						 // successful exit
}     

// int WriteFlashToFlash(const unsigned long FlashInAddress, const unsigned long FlashOutAddress, const unsigned long FlashLength)
// Read a section of Flash and write image to another flash
//
//	FlashInAddress	= Source Memory address to read from
//	FlashOutAddress	= Destination Memory address to write to
//	FlashLength	= Number of bytes to write to the image file from Flash
//
// Returns 0 if no error, non-zero if any errors
//
int WriteFlashToFlash(const unsigned long FlashInAddress, const unsigned long FlashOutAddress, const unsigned long FlashLength)
{
	char  Buffer[BUFFER_SIZE];	// data buffer
//	int 	i = 0;			// loop count
//	int	out_file;		// output file descriptor
	unsigned long  BytesWritten;	 // number of bytes written to the file
	unsigned long  BytesTotal = 0;	// total bytes written to the file
	unsigned long  FlashInTemp = FlashInAddress;	// various temp. large numbers
	unsigned long  FlashOutTemp = FlashOutAddress;	// various temp. large numbers
	unsigned long  read_size = 0;	// bytes read in last read
//	unsigned long	read_total = 0;	// running total of bytes


//	CheckModel();

	cout << "Buffer Address = " << hex << FP_SEG(Buffer) << ":" << hex << FP_OFF(Buffer) << endl;
	cout << "Flash source address: 0x" << hex << FlashInAddress << endl;
	cout << "Flash destination address: 0x" << hex << FlashOutAddress << endl;
	cout << "Flash length:  0x" << hex << FlashLength << endl;


	// erase the Flash memory
	for(FlashOutTemp = FlashOutAddress; FlashOutTemp < (FlashOutAddress + FlashLength); FlashOutTemp += 0x10000)
		{

		cout << "\rErasing Flash memory: Address = 0x" << hex << FlashOutTemp;

		if((EraseFlash (FlashOutTemp)) != 0)
			{
			cout << endl << endl << "Flash memory erase command failure, exiting." << endl;
			return(ERR_FLASH);
			}
		}
	cout << "\rErasing Flash memory: complete                           " << endl;

	for(FlashOutTemp = FlashOutAddress; FlashOutTemp < (FlashOutAddress + FlashLength); FlashOutTemp += BUFFER_SIZE)
		{
		cout << "\rReading Flash memory: address = 0x" << hex << FlashInTemp
		<< ", Bytes = 0x" << hex << BUFFER_SIZE;

		if(ReadFlash (FlashInTemp, FP_SEG(Buffer), FP_OFF(Buffer), BUFFER_SIZE))
			{
			cerr << "Error: cannot read from Flash memory - Unknown Error" << endl;
			return(ERR_FLASH);
			}

		cout << "\rWriting Flash memory: address = 0x" << hex << FlashOutTemp
		<< ", Bytes = 0x" << hex << read_size;

		if((WriteFlash(FlashOutTemp, FP_SEG(Buffer), FP_OFF(Buffer), BUFFER_SIZE)) != 0)
			{
			cout << endl << endl << "Flash memory write command failure, exiting." << endl;
			return(ERR_FLASH);
			}

		FlashInTemp += BUFFER_SIZE;
		BytesTotal += BUFFER_SIZE;

		}

	// All finished. Close things down and report success
	cout << "\rCopying Flash memory: complete                              " << endl; 
	cout << "Flash image successfully copied. Total bytes = 0x" << hex << BytesTotal << endl;

	return(ERR_NONE);						 // successful exit
}     

int main(int argc, char *argv[])
{
	char  Flag;							// confirmation response
	int   i;								// temp loop control
	int   ModelNum = 0;				// model number of current board
	int   Error = 0;					// Error code holder
	char  StrTmp[256] = "";	// holder for cmd line arguments
	char  FileDisk[20] = "";		// filename holder
	char  FileDos[20] = "";	// filename holder
	char  FileBios[20] = "";	// filename holder
	char  FileCmos[20] = "";	// filename holder
	char  *Ptr;
	int   DiskFlag = 0;				// read/write disk image
	int   DosFlag = 0;				// r/w dos image
	int   BiosFlag = 0;				// r/w bios image
	int   CmosFlag = 0;				// r/w bios image
	int   RtcFlag = 0;		// update RTC from TS-9200
	int   WriteBoot = 0;		// write TO Bootboard (TS-91/9200) if set
	int   WriteFlash = 0;			// write TO Flash if set
	int   WriteFile = 0;				// write TO image files if set
	char  FilePrefix[8] = "";	// filename prefix (optional)
//	int	FilePre = 0;		// filename prefix flag
	int   EepromXfer = 0;		// transfer 1000:0000 <-> 2000:0000 if set
//	unsigned long MBase = 0;	// Memory base 

	unsigned char CmosData[128] = {0xFF};


	cout << VER << endl << endl;	// print program version

	if(argc < 2)
		{
		Usage();		// invalid command line - display usage and exit
		}

	for( i = 1; i < argc ; ++i )
		{
		strncpy(StrTmp, argv[i], 255);

		if( !stricmp( StrTmp, "/BLAST") || !stricmp( StrTmp,"-BLAST"))
			{
			WriteFlash = 1;

			}
		else if( !stricmp( StrTmp, "/CREATE") || !stricmp( StrTmp,"-CREATE"))
			{
			WriteFile = 1;

			}
		else if( !stricmp( StrTmp, "/EEPROM") || !stricmp( StrTmp,"-EEPROM"))
			{
			EepromXfer = 1;

			}
		else if( !stricmp( StrTmp, "/BLASTBOOT") || !stricmp( StrTmp,"-BLASTBOOT"))
			{
			WriteBoot = 1;

			}
		else if(!strncmpi(StrTmp, "DISK", 4))
			{
			DiskFlag = 1;
			if((StrTmp[4] == '=') && (StrTmp[5] != '\0'))
				{
				strncpy(FileDisk, &StrTmp[5], 12);
				FileDisk[13] = '\0';
				}

			}
		else if(!strncmpi(StrTmp, "DOS", 3))
			{
			DosFlag = 1;
			if((StrTmp[3] == '=') && (StrTmp[4] != '\0'))
				{
				strncpy(FileDos, &StrTmp[4], 12);
				FileDisk[13] = '\0';
				}

			}
		else if( !strncmpi(StrTmp, "BIOS", 4))
			{
			BiosFlag = 1;
			if((StrTmp[4] == '=') && (StrTmp[5] != '\0'))
				{
				strncpy(FileBios, &StrTmp[5], 12);
				FileDisk[13] = '\0';
				}

			}
		else if( !strncmpi(StrTmp, "CMOS", 4))
			{
			CmosFlag = 1;
			if((StrTmp[4] == '=') && (StrTmp[5] != '\0'))
				{
				strncpy(FileCmos, &StrTmp[5], 12);
				FileCmos[13] = '\0';
				}

			}
		else if( !strncmpi(StrTmp, "RTC", 3))
			{
			RtcFlag = 1;

			}
		else if( !strncmpi(StrTmp, "ALL", 3))
			{
			DiskFlag = 1;
			DosFlag = 1;
			BiosFlag = 1;
			CmosFlag = 1;
			RtcFlag = 1;

			}
		else
			{
			Usage();							// bad command flag / help
			}
		}

	ModelNum = CheckModel();		  // find and display EPC model number


	itoa(ModelNum, FilePrefix, 10);

	if(FileDisk[0] == '\0')
		{
		strcpy(FileDisk, FilePrefix);
		strncat(FileDisk, "DISK.BIN", 11);
		}
	if(FileDos[0] == '\0')
		{
		strcpy(FileDos, FilePrefix);
		strncat(FileDos, "DOS.BIN", 11);
		}
	if(FileBios[0] == '\0')
		{
		strcpy(FileBios, FilePrefix);
		strncat(FileBios, "BIOS.BIN", 11);

		}
	if(FileCmos[0] == '\0')
		{
		strcpy(FileCmos, FilePrefix);
		strncat(FileCmos, "CMOS.BIN", 11);

		}


// 	exit(0);

	outp (0x472, 0x01);				  // RED LED ON
	outp (0x473, 0x00);				  // GREEN LED OFF

	if(EepromXfer)
		{
		cout << "Preparing to update the EPC Flash memory from EEPROM." << endl;
		cout << "WARNING: ALL CURRENT CONTENTS OF THE EPC's: ";
		if(BiosFlag) cout << "BIOS ";
		if(DosFlag)	cout << "DOS ";
		if(DiskFlag) cout << "FLASHDISK ";
		if(CmosFlag) cout << "CMOS ";
		if(RtcFlag)	cout << "RTC ";
		cout << " WILL BE LOST!" << endl << endl;
		cout << "Continue? (Y/N) ";
		Flag = getch();				// get response. getche() doesn't appear to work..

		if( Flag != 'Y' && Flag != 'y')
			{
			exit(ERR_USAGE);
			}

		if(BiosFlag)
			{
			cout << "Writing EEPROM BIOS image to Flash memory:\n";
			Error = WriteFlashToFlash(EEPROMBASE + BIOSOFFSET, FLASHBASE + BIOSOFFSET, 0x10000);
			if(Error) ErrorHandler(Error);
			}
		if(DosFlag)
			{
			cout << "Writing EEPROM DOS image to Flash memory:\n";
			Error = WriteFlashToFlash(EEPROMBASE + DOSOFFSET,  FLASHBASE + DOSOFFSET,  0x10000);
			if(Error) ErrorHandler(Error);
			}
		if(DiskFlag)
			{
			cout << "Writing EEPROM disk image file to Flash memory:\n";
			Error = WriteFlashToFlash(EEPROMBASE + DISKOFFSET, FLASHBASE + DISKOFFSET, 0xE0000);
			if(Error) ErrorHandler(Error);
			}

		}										  // if (WriteFlash)

	else if(WriteFlash)
		{
		cout << "Preparing to update the EPC." << endl;
		cout << "WARNING: ALL CURRENT CONTENTS OF THE EPC's: ";
		if(BiosFlag) cout << "BIOS ";
		if(DosFlag)	cout << "DOS ";
		if(DiskFlag) cout << "FLASHDISK ";
		if(CmosFlag) cout << "CMOS ";
		if(RtcFlag)	cout << "RTC ";
		cout << " WILL BE LOST!" << endl << endl;
		cout << "Continue? (Y/N) ";
		Flag = getch();				// get response. getche() doesn't appear to work..
		if( Flag != 'Y' && Flag != 'y')
			{
			exit(ERR_USAGE);
			}

		// CMOS stuff
		if(CmosFlag)
			{
			cout << "Writing file to EPC CMOS: " << FileCmos << endl;
			Error = WriteFileToCmos(FileCmos, CmosData);
			if(Error) ErrorHandler(Error);
			}

		// RTC stuff
		//
		// This only works on a TS9200, but we want to use the same batch files that
		// call FLASHUP on a TS9100. So only crap out if the error is other than no9200.
		if(RtcFlag)
			{
			cout << "Updating RTC contents: " << endl;
			Error = UpdateRtc();
			if(Error && (Error != ERR_NO9200)) ErrorHandler(Error);
			}


		if(BiosFlag)
			{
			cout << "Writing BIOS image file " << FileBios << " to Flash memory:\n";
			Error = WriteFileToFlash(FileBios, FLASHBASE + BIOSOFFSET, 0x10000);
			if(Error) ErrorHandler(Error);
			}
		if(DosFlag)
			{
			cout << "Writing DOS image file " << FileDos << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDos,  FLASHBASE + DOSOFFSET,  0x10000);
			if(Error) ErrorHandler(Error);
			}
		if(DiskFlag)
			{
			cout << "Writing disk image file " << FileDisk << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDisk, FLASHBASE + DISKOFFSET, 0xE0000);
			if(Error) ErrorHandler(Error);
			}

		}										  // if (WriteFlash)
	else if(WriteFile)
		{
		// CMOS stuff
		if(CmosFlag)
			{
			cout << "Writing CMOS to file " << FileCmos << endl;
			Error = WriteCmosToFile(FileCmos, CmosData);
			if(Error) ErrorHandler(Error);
			}

		if(BiosFlag)
			{
			cout << "Writing BIOS to image file " << FileBios << ":\n";
			Error = WriteFlashToFile(FileBios, FLASHBASE + BIOSOFFSET, 0x10000);
			if(Error) ErrorHandler(Error);
			}
		if(DosFlag)
			{
			cout << "Writing DOS to image file " << FileDos << ":\n";
			Error = WriteFlashToFile(FileDos,  FLASHBASE + DOSOFFSET,  0x10000);
			if(Error) ErrorHandler(Error);
			}
		if(DiskFlag)
			{
			cout << "Writing disk image to file " << FileDisk << ":\n";
			Error = WriteFlashToFile(FileDisk, FLASHBASE + DISKOFFSET, 0xE0000);
			if(Error) ErrorHandler(Error);
			}

		}							 // if (WriteFile)

	else if(WriteBoot)
		{
		cout << "Preparing to update the BOOT BOARD FLASH" << endl;
		cout << "WARNING: CURRENT CONTENTS OF THE BOOT BOARD'S: ";
		if(BiosFlag) cout << "BIOS ";
		if(DosFlag)	cout << "DOS ";
		if(DiskFlag) cout << "FLASHDISK ";
//		if (CmosFlag) cout << "CMOS ";
//		if (RtcFlag) cout << "RTC ";
		cout << " WILL BE LOST!" << endl << endl;
		cout << "Continue? (Y/N) ";
		Flag = getch();				// get response. getche() doesn't appear to work..
		if( Flag != 'Y' && Flag != 'y')
			{
			exit(ERR_USAGE);
			}

//		// CMOS stuff
//		if (CmosFlag) {
//			cout << "Writing file to EPC CMOS: " << FileCmos << endl;
//			Error = WriteFileToCmos(FileCmos, CmosData);
//			if (Error) ErrorHandler(Error);
//		}

//		// RTC stuff
//		//
//		// This only works on a TS9200, but we want to use the same batch files that
//		// call FLASHUP on a TS9100. So only crap out if the error is other than no9200.
//		if (RtcFlag) {
//			cout << "Updating RTC contents: " << endl;
//			Error = UpdateRtc();
//			if (Error && (Error != ERR_NO9200)) ErrorHandler(Error);
//		}


		if(BiosFlag)
			{
			cout << "Writing BIOS image file " << FileBios << " to Flash memory:\n";
			Error = WriteFileToFlash(FileBios, BOOTBASE + BOOTBIOSOFFSET, 0x10000);
			if(Error) ErrorHandler(Error);
			}
		if(DosFlag)
			{
			cout << "Writing DOS image file " << FileDos << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDos,  BOOTBASE + BOOTDOSOFFSET,  0x10000);
			if(Error) ErrorHandler(Error);
			}
		if(DiskFlag)
			{
			cout << "Writing disk image file " << FileDisk << " to Flash memory:\n";
			Error = WriteFileToFlash(FileDisk, BOOTBASE + BOOTDISKOFFSET, 0x60000);
			if(Error) ErrorHandler(Error);
			}

		}										  // if (WriteFlash)

	outp (0x472, 0x00);				  // RED LED OFF
	outp (0x473, 0x01);				  // GREEN LED ON

	if(WriteFlash)
		{
		cout << "\nDone. Turn off power, remove the boot board, and turn the power back on." << endl;
		}

	return(ERR_NONE);
}
